/*********************************************************************
*               (c) SEGGER Microcontroller GmbH & Co. KG             *
*                        The Embedded Experts                        *
*                           www.segger.com                           *
**********************************************************************
----------------------------------------------------------------------
File    : FlashPrg.c
Purpose : Implementation of RAMCode template
--------  END-OF-HEADER  ---------------------------------------------
*/
#include "FlashOS.H"

#include "main.h"
#include "stm32f4xx_hal.h"

#include "mx25.h"

/*********************************************************************
*
*       Defines (configurable)
*
**********************************************************************
*/
#define PAGE_SIZE_SHIFT (8) // The smallest program unit (one page) is 256 byte in size
//
// Some flash types require a native verify function as the memory is not memory mapped available (e.g. eMMC flashes).
// If the verify function is implemented in the algorithm, it will be used by the J-Link DLL during compare / verify
// independent of what verify type is configured in the J-Link DLL.
// Please note, that SEGGER does not recommend to use this function if the flash can be memory mapped read
// as this may can slow-down the compare / verify step.
//
#define SUPPORT_NATIVE_VERIFY (0)
#define SUPPORT_NATIVE_READ_BACK (1)
#define SUPPORT_BLANK_CHECK (0)

/*********************************************************************
*
*       Types
*
**********************************************************************
*/

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/
//
// We use this dummy variable to make sure that the PrgData
// section is present in the output elf-file as this section
// is mandatory in current versions of the J-Link DLL
//
static volatile int _Dummy;

/*********************************************************************
*
*       Static code
*
**********************************************************************
*/

/*********************************************************************
*
*       _FeedWatchdog
*
*  Function description
*    Feeds the watchdog. Needs to be called during RAMCode execution
*    in case of an watchdog is active.
*/
static void _FeedWatchdog(void)
{
}

/*********************************************************************
*
*       Public code
*
**********************************************************************
*/

/*********************************************************************
*
*       Init
*
*  Function description
*    Handles the initialization of the flash module.
*
*  Parameters
*    Addr: Flash base address
*    Freq: Clock frequency in Hz
*    Func: Caller type (e.g.: 1 - Erase, 2 - Program, 3 - Verify)
*
*  Return value 
*    0 O.K.
*    1 Error
*/
int Init(U32 Addr, U32 Freq, U32 Func)
{
    (void)Addr;
    (void)Freq;
    (void)Func;
    int ret;

    //CSP init; clocks & IO & peripherals
    ret = platform_init_sequence();

    // Read MFG ID
    uint8_t mfg_id;
    uint16_t device_id;
    HAL_StatusTypeDef hal_status;
    hal_status = (HAL_StatusTypeDef)mx25_read_mfg_device_id( &mfg_id, &device_id );
    if (hal_status != HAL_OK)
    {
        return 1;
    }

    for (int i; i < 8; i++)
    {
        uint8_t flash_status;

        hal_status = mx25_read_status( &flash_status );
        if (hal_status != HAL_OK)
        {
            continue;
        }
        //check if status indicates ready
        // if ready, return 0
        if ( !(flash_status & 0x01) ){ // check if WIP is 0 
          return 0;
        }
    }

    return 1;
}

/*********************************************************************
*
*       UnInit
*
*  Function description
*    Handles the de-initialization of the flash module.
*
*  Parameters
*    Func: Caller type (e.g.: 1 - Erase, 2 - Program, 3 - Verify)
*
*  Return value 
*    0 O.K.
*    1 Error
*/
int UnInit(U32 Func)
{
    (void)Func;
    //
    // Uninit code
    //
    *(volatile U32 *)(0x1234) = Func; // Dummy
    return 0;
}

/*********************************************************************
*
*       EraseSector
*
*  Function description
*    Erases one flash sector.
*
*  Parameters
*    Addr: Address of the sector to be erased
*
*  Return value 
*    0 O.K.
*    1 Error
*/
int EraseSector(U32 SectorAddr)
{
    int r;

    // This assumes that the passed address is the first address of this sector OR an address within the sector
    r = mx25_erase_sector(SectorAddr);

    _FeedWatchdog();
    return r;
}

/*********************************************************************
*
*       ProgramPage
*
*  Function description
*    Programs one flash page.
*
*  Parameters
*    DestAddr: Destination address
*    NumBytes: Number of bytes to be programmed (always a multiple of program page size, defined in FlashDev.c)
*    pSrcBuff: Point to the source buffer
*
*  Return value 
*    0 O.K.
*    1 Error
*/
int ProgramPage(U32 DestAddr, U32 NumBytes, U8 *pSrcBuff)
{
//    volatile U8 *pSrc;
//    volatile U8 *pDest;
//    U8 AccessWidth;
//    U32 Status;
    U32 NumPages;
    U32 NumBytesAtOnce;
    int status = -1;
    U32 pos = 0;

//    r = -1;
//    pSrc = (volatile U8 *)pSrcBuff;
//    pDest = (volatile U8 *)DestAddr;
    //
    // RAMCode is able to program multiple pages
    //
    NumPages = NumBytes >> PAGE_SIZE_SHIFT;
    //
    // Program page-wise
    //
    if (NumPages)
    {   
        NumBytesAtOnce = (1 << PAGE_SIZE_SHIFT);
        _FeedWatchdog();
        //
        // Program one page
        //
        do
        {
            //
            // Program page code
            status = mx25_prog_page( DestAddr + pos, NumBytesAtOnce, &(pSrcBuff[pos]) );
            if( status != 0){ return status; }
            //
//                *pDest++ = *pSrc++;
            pos += NumBytesAtOnce;
        } while (--NumPages);
    }
//    return status;
    return 0;
}

/*********************************************************************
*
*       Verify
*
*  Function description
*    Compares a specified number of bytes of a provided data
*    buffer with the content of the device
*
*  Parameters
*    Addr: Start address in memory which should be compared
*    NumBytes: Number of bytes to be compared
*    pBuff: Pointer to the data to be compared
*
*  Return value 
*    == (Addr + NumBytes): O.K.
*    != (Addr + NumBytes): *not* O.K.
*
*/
#if SUPPORT_NATIVE_VERIFY
U32 Verify(U32 Addr, U32 NumBytes, U8 *pBuff)
{
    unsigned char *pFlash;
    unsigned long r;

    pFlash = (unsigned char *)Addr;
    r = Addr + NumBytes;
    do
    {
        if (*pFlash != *pBuff)
        {
            r = (unsigned long)pFlash;
            break;
        }
        pFlash++;
        pBuff++;
    } while (--NumBytes);
    return r;
}
#endif

/*********************************************************************
*
*       BlankCheck
*
*  Function description
*    Checks if a memory region is blank
*
*  Parameters
*    Addr: Blank check start address
*    NumBytes: Number of bytes to be checked
*    BlankData: Pointer to the destination data
*
*  Return value 
*    0 O.K., blank
*    1 O.K., *not* blank
*    <  0 Error
*
*/
#if SUPPORT_BLANK_CHECK
int BlankCheck(U32 Addr, U32 NumBytes, U8 BlankData)
{
    U8 *pData;

    pData = (U8 *)Addr;
    do
    {
        if (*pData++ != BlankData)
        {
            return 1;
        }
    } while (--NumBytes);
    return 0;
}
#endif

/*********************************************************************
*
*       SEGGER_OPEN_Read
*
*  Function description
*    Reads a specified number of bytes into the provided buffer
*
*  Parameters
*    Addr: Start read address
*    NumBytes: Number of bytes to be read
*    pBuff: Pointer to the destination data
*
*  Return value 
*    >= 0 O.K., NumBytes read
*    <  0 Error
*
*/
#if SUPPORT_NATIVE_READ_BACK

#define min(a,b) \
   ({ __typeof__ (a) _a = (a); \
       __typeof__ (b) _b = (b); \
     _a < _b ? _a : _b; })

int SEGGER_OPEN_Read(U32 Addr, U32 NumBytes, U8 *pDestBuff)
{
    //
    // Read function
    int status;
    U32 pos = 0;
    while(  pos < NumBytes ){
      U32 bytes = min( NumBytes - pos , 1024);
      
      status = mx25_read( Addr + pos, bytes, &(pDestBuff[pos]) );
      if( status != 0){ return status; }

      pos += bytes;
    }


    return NumBytes;
}
#endif
