#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include "main.h"

#include "mx25.h"

#include "stm32f4xx.h"
#include "stm32f4xx_hal.h"
#include "stm32f4xx_hal_spi.h"

#define SFLASH_READ_TIMEOUT_MS  (1000)

// Taken from WICED SDK
/* Command definitions */
typedef enum
{
    SFLASH_WRITE_STATUS_REGISTER        = 0x01, /* WRSR                   */
    SFLASH_WRITE                        = 0x02,
    SFLASH_READ                         = 0x03,
    SFLASH_WRITE_DISABLE                = 0x04, /* WRDI                   */
    SFLASH_READ_STATUS_REGISTER         = 0x05, /* RDSR                   */
    SFLASH_WRITE_ENABLE                 = 0x06, /* WREN                   */
    SFLASH_FAST_READ                    = 0x0B,
    SFLASH_SECTOR_ERASE                 = 0x20, /* SE                     */
    SFLASH_READ_STATUS_REGISTER2        = 0x35, /* RDSR-2 - Winbond only  */
    SFLASH_BLOCK_ERASE_MID              = 0x52, /* SE                     */
    SFLASH_BLOCK_ERASE_LARGE            = 0xD8, /* SE                     */
    SFLASH_READ_ID1                     = 0x90, /* data size varies       */
    SFLASH_READ_ID2                     = 0xAB, /* data size varies       */
    SFLASH_READ_JEDEC_ID                = 0x9F, /* RDID                   */
    SFLASH_CHIP_ERASE1                  = 0x60, /* CE                     */
    SFLASH_CHIP_ERASE2                  = 0xC7, /* CE                     */
    SFLASH_ENABLE_WRITE_STATUS_REGISTER = 0x50, /* EWSR   - SST only      */
    SFLASH_READ_SECURITY_REGISTER       = 0x2B, /* RDSCUR - Macronix only */
    SFLASH_WRITE_SECURITY_REGISTER      = 0x2F, /* WRSCUR - Macronix only */
    SFLASH_ENTER_SECURED_OTP            = 0xB1, /* ENSO   - Macronix only */
    SFLASH_EXIT_SECURED_OTP             = 0xC1, /* EXSO   - Macronix only */
    SFLASH_DEEP_POWER_DOWN              = 0xB9, /* DP     - Macronix only */
    SFLASH_RELEASE_DEEP_POWER_DOWN      = 0xAB, /* RDP    - Macronix only */
    SFLASH_QUAD_READ                    = 0x6B, /* QREAD  - Macronix only */
    SFLASH_QUAD_WRITE                   = 0x32, /* QWRITE - ISSI only     */
    SFLASH_X4IO_WRITE                   = 0x38, /* 4PP    - Macronix only */
    SFLASH_X4IO_READ                    = 0xEB, /* 4READ  - Macronix only */
    SFLASH_DIE_ERASE_MICRON             = 0xC4, /* Die Erase - Micron only */
    SFLASH_SECTOR_ERASE_MICRON          = 0xD8, /* Sector Erase (64KB) - Micron only */
    SFLASH_CLEAR_FLAG_STATUS_REGISTER   = 0x50,  /* Micro Only */
    SFLASH_READ_FLAG_STATUS_REGISTER    = 0x70,  /* Micron Only */
    SFLASH_WRITE_ENH_VOLATILE_REGISTER  = 0x61, /* WRITE Enhance status register, Micron only */
    SFLASH_READ_ENH_VOLATILE_REGISTER   = 0x65, /* READ Enhance status register, Micron only */
} sflash_command_t;

extern SPI_HandleTypeDef hspi1;

#define SFLASH_READ_MFG_ID_TIMEOUT_MS   50
#define SFLASH_READ_STATUS_TIMEOUT_MS   50
#define SFLASH_COMMAND_TIMEOUT_MS       50
#define SFLASH_WRITE_ENABLE_TIMEOUT_MS  50
#define SFLASH_PROG_PAGE_TIMEOUT_MS     100

// Local functions

static HAL_StatusTypeDef wait_for_ready( int32_t max_retries, uint32_t ms_pause );
static HAL_StatusTypeDef send_write_enable( void );

static inline void sflash_CS_HIGH( void ) {
  HAL_GPIO_WritePin(SFLASH_NSS_GPIO_Port, SFLASH_NSS_Pin, GPIO_PIN_SET);
}

static inline void sflash_CS_LOW( void ) {
  HAL_GPIO_WritePin(SFLASH_NSS_GPIO_Port, SFLASH_NSS_Pin, GPIO_PIN_RESET);
}

// Public functions

int mx25_read_mfg_device_id (uint8_t *mfg_id, uint16_t* device_id )
{
  HAL_StatusTypeDef status;
  uint8_t tx_data[] = { SFLASH_READ_JEDEC_ID, 0, 0, 0 }; //command, 3B dummy
  uint8_t rx_data[ sizeof(tx_data) ] = {0,};
  
  sflash_CS_LOW();
  status = HAL_SPI_TransmitReceive( &hspi1, tx_data, rx_data, sizeof(tx_data), SFLASH_READ_MFG_ID_TIMEOUT_MS );
  sflash_CS_HIGH();

  if( status == HAL_OK ){
    *mfg_id    = rx_data[1];
    *device_id = (rx_data[2] << 8) + rx_data[3]; 
  }

  __HAL_SPI_DISABLE(&hspi1);
 
  return (int)status;
}

int mx25_read_status(uint8_t *flash_status)
{
  HAL_StatusTypeDef status;
  uint8_t tx_data[2] = { SFLASH_READ_STATUS_REGISTER, 0};
  uint8_t rx_data[2] = {0,};

  sflash_CS_LOW();
  status = HAL_SPI_TransmitReceive( &hspi1, tx_data, rx_data, 2, SFLASH_READ_STATUS_TIMEOUT_MS );
  sflash_CS_HIGH();

  if( status == HAL_OK ){
    *flash_status = rx_data[1];
  }

  __HAL_SPI_DISABLE(&hspi1);

  return status;
}

// Address should be any address within the sector to be erased
int mx25_erase_sector(uint32_t addr)
{
  HAL_StatusTypeDef status;

  status = send_write_enable();
  if( status != HAL_OK ){
    return (int)status;
  }

  status = wait_for_ready(8, 10);
  if( status != HAL_OK ){
    return (int)status;
  }

  //send erase sector command, with address
  uint8_t tx_data[4];
  tx_data[0] = SFLASH_SECTOR_ERASE;
  tx_data[1] = (uint8_t)(addr & 0x0000FF);
  tx_data[2] = (uint8_t)((addr & 0x00FF00) >> 8);
  tx_data[3] = (uint8_t)((addr & 0xFF0000) >> 16);
 
  sflash_CS_LOW();
  status = HAL_SPI_Transmit( &hspi1, tx_data, sizeof(tx_data), SFLASH_COMMAND_TIMEOUT_MS );
  sflash_CS_HIGH();
  __HAL_SPI_DISABLE(&hspi1);

  if( status != HAL_OK ){
    return (int)status;
  }

  //wait with no limit on retries, until chip erase is compelted
  status = wait_for_ready(-1, 10);

  return status;
}

int mx25_erase_chip(void)
{
 HAL_StatusTypeDef status;

  status = send_write_enable(); //send write enable (?)
  if( status != HAL_OK ){
    return (int)status;
  }

  status = wait_for_ready(8, 10);
  if( status != HAL_OK ){
    return (int)status;
  }

  //send erase chip command
  uint8_t tx_data[] = { SFLASH_CHIP_ERASE1 };
  sflash_CS_LOW();
  status = HAL_SPI_Transmit( &hspi1, tx_data, sizeof(tx_data), SFLASH_COMMAND_TIMEOUT_MS ); 
  sflash_CS_HIGH();
  if( status != HAL_OK ){
    return (int)status;
  }

  //wait with no limit on retries, until chip erase is compelted
  status = wait_for_ready(-1, 10);

  return status;
}

int mx25_read(uint32_t addr, uint32_t num, uint8_t *data)
{
  //wait for status ready
  uint8_t flash_status;
  HAL_StatusTypeDef status;
  uint8_t retries = 0;

  do{
    // Up to 8 retires to wait for write ready
    if(retries > 8){
      return 1;
    }
    status = mx25_read_status( &flash_status );
  } while( (status != HAL_OK) || (flash_status & 0x01) );

  //send read command
  uint8_t tx_data[4];
  tx_data[0] = SFLASH_READ;
  tx_data[1] = (uint8_t)(addr & 0x0000FF);
  tx_data[2] = (uint8_t)((addr & 0x00FF00) >> 8);
  tx_data[3] = (uint8_t)((addr & 0xFF0000) >> 16);

  //send command and address
  sflash_CS_LOW();
  status = HAL_SPI_Transmit( &hspi1, tx_data, sizeof(tx_data), SFLASH_READ_TIMEOUT_MS );
  if( status != HAL_OK ){
    sflash_CS_HIGH();
    __HAL_SPI_DISABLE(&hspi1);
   return (int)status;
  }

  // Read back data directly into output pointer;
  status = HAL_SPI_Receive( &hspi1, data, num, SFLASH_READ_TIMEOUT_MS );
  sflash_CS_HIGH();

  __HAL_SPI_DISABLE(&hspi1);
  return (int)status;
}


int mx25_prog_page(uint32_t addr, uint32_t num, uint8_t* data)
{
  uint8_t flash_status;
  HAL_StatusTypeDef status = 0;
  uint8_t retries = 0;

  do{
    // Up to 8 retires to wait for write ready
    if(retries > 8){
      return 1;
    }
    status = mx25_read_status( &flash_status );
  } while( (status != HAL_OK) || (flash_status & 0x01) );

  status = send_write_enable();  //send write enable
  if( status != HAL_OK ){
    return (int)status;
  }

  uint8_t tx_data[256 + 4];
//  uint8_t rx_data[4];

  memset(tx_data, 0, sizeof(tx_data));
//  memset(rx_data, 0, sizeof(rx_data));

  // Set up TX data:
  tx_data[0] = SFLASH_WRITE;
  tx_data[1] = (uint8_t)(addr & 0x0000FF);
  tx_data[2] = (uint8_t)((addr & 0x00FF00) >> 8);
  tx_data[3] = (uint8_t)((addr & 0xFF0000) >> 16);

  memcpy( &tx_data[4], data, num );
  uint32_t trans_len = num + 1 + 3; //1 command, 3 address, num data

  //Do transaction
  sflash_CS_LOW();
  status = HAL_SPI_Transmit( &hspi1, tx_data, trans_len, SFLASH_PROG_PAGE_TIMEOUT_MS );
  __HAL_SPI_DISABLE(&hspi1);
  if( status != HAL_OK ){
    return 1;
  }


  //poll status until write is done
  do{
    status = mx25_read_status( &flash_status );
  } while( flash_status & 0x01 );

  // Copy data to output
//  memcpy(data, &(rx_data[3]), num );

  return (int)status;

}

static HAL_StatusTypeDef send_write_enable(void)
{
  HAL_StatusTypeDef status;
  uint8_t data = SFLASH_WRITE_ENABLE;
  sflash_CS_LOW();
  status = HAL_SPI_Transmit(&hspi1, &data, 1,  SFLASH_WRITE_ENABLE_TIMEOUT_MS);
  sflash_CS_LOW();
  
  __HAL_SPI_DISABLE(&hspi1);
  return status;
}

static HAL_StatusTypeDef wait_for_ready(int32_t max_retries, uint32_t ms_pause)
{
  uint8_t flash_status;
  HAL_StatusTypeDef status;
  uint32_t retries = 0;

  do{
    status = mx25_read_status( &flash_status );
    retries++;
    if( (max_retries > 0) && (retries > max_retries) ){
      return status;
    }
    //pause for a short while
    for( volatile int counter = 0; counter < 1000; counter ++) {
      asm("NOP");
    }
    
  } while( (status != HAL_OK) || (flash_status & 0x01) );

  return status;
}